@@ -6,19 +6,20 @@ if int(version("shapely").split(".")[0]) > 1:
 
     os.environ["USE_PYGEOS"] = "0"
 
-from ._utils import transform2tuple
-from .exceptions import (
+from pygeoutils._utils import get_gtiff_attrs, transform2tuple
+from pygeoutils.exceptions import (
     EmptyResponseError,
     InputRangeError,
     InputTypeError,
     InputValueError,
+    MatchingCRSError,
     MissingAttributeError,
     MissingColumnError,
     MissingCRSError,
     UnprojectedCRSError,
 )
-from .print_versions import show_versions
-from .pygeoutils import (
+from pygeoutils.print_versions import show_versions
+from pygeoutils.pygeoutils import (
     Coordinates,
     GeoBSpline,
     arcgis2geojson,
@@ -26,11 +27,11 @@ from .pygeoutils import (
     geo2polygon,
     geodf2xarray,
     geometry_list,
-    get_gtiff_attrs,
     get_transform,
     gtiff2xarray,
     json2geodf,
     nested_polygons,
+    query_indices,
     snap2nearest,
     xarray2geodf,
     xarray_geomask,
@@ -57,6 +58,7 @@ __all__ = [
     "xarray_geomask",
     "Coordinates",
     "GeoBSpline",
+    "query_indices",
     "nested_polygons",
     "InputTypeError",
     "InputValueError",
@@ -64,6 +66,7 @@ __all__ = [
     "MissingAttributeError",
     "MissingColumnError",
     "MissingCRSError",
+    "MatchingCRSError",
     "UnprojectedCRSError",
     "EmptyResponseError",
     "show_versions",
