@@ -2,41 +2,23 @@
 from __future__ import annotations
 
 import itertools
-import os
-import sys
 from dataclasses import dataclass
 from numbers import Number
-from typing import Any, NamedTuple, Union
+from typing import TYPE_CHECKING, Any, NamedTuple, TypeVar, Union, cast
 
 import numpy as np
 import pyproj
 import rasterio as rio
+import rioxarray._io as rxr
 import ujson as json
 import xarray as xr
-from loguru import logger
 from shapely.geometry import LineString, Point
 
-NUMBER = Union[int, float, np.number]
+from pygeoutils.exceptions import MissingAttributeError
 
-logger.configure(
-    handlers=[
-        {
-            "sink": sys.stdout,
-            "colorize": True,
-            "format": " | ".join(
-                [
-                    "{time:YYYY-MM-DD at HH:mm:ss}",  # noqa: FS003
-                    "{name: ^15}.{function: ^15}:{line: >3}",  # noqa: FS003
-                    "{message}",  # noqa: FS003
-                ]
-            ),
-        }
-    ]
-)
-if os.environ.get("HYRIVER_VERBOSE", "false").lower() == "true":
-    logger.enable("pygeoutils")
-else:
-    logger.disable("pygeoutils")
+if TYPE_CHECKING:
+    NUMBER = Union[int, float, np.number]  # type: ignore
+    XD = TypeVar("XD", xr.Dataset, xr.DataArray)
 
 
 @dataclass
@@ -130,15 +112,11 @@ class Convert:
                 "curveRings": "Curved Polygon",
                 "curvePaths": "Curved Polyline",
                 "a": "Elliptic Arc",
-                "b": "Bézier Curve",
+                "b": "Bezier Curve",
                 "c": "Circular Arc",
             }
             not_supported = [v for k, v in curves.items() if k in arcgis["geometry"]]
             if not_supported:
-                logger.warning(
-                    "Elements of type {} can not be converted. Converting to null geometry",
-                    ", ".join(not_supported),
-                )
                 geojson["geometry"] = None
             else:
                 geojson["geometry"] = convert(arcgis["geometry"])
@@ -250,15 +228,16 @@ class Attrs(NamedTuple):
     transform: tuple[float, float, float, float, float, float]
 
 
-def get_nodata(src: Any) -> NUMBER:
+def get_nodata(src: Any) -> np.number:  # type: ignore
     """Get the nodata value of a GTiff byte response."""
     if src.nodata is None:
         try:
-            nodata: NUMBER = np.iinfo(src.dtypes[0]).max
+            nodata = np.iinfo(src.dtypes[0]).max
         except ValueError:
             nodata = np.nan
     else:
         nodata = np.dtype(src.dtypes[0]).type(src.nodata)
+    nodata = cast("np.number", nodata)  # type: ignore
     return nodata
 
 
@@ -268,7 +247,9 @@ def get_dim_names(ds: xr.DataArray | xr.Dataset) -> tuple[str, str] | None:
     x_dims = {"x", "X", "lon", "Lon", "longitude", "Longitude"}
     try:
         y_dim = list(set(ds.coords).intersection(y_dims))[0]
+        y_dim = cast("str", y_dim)
         x_dim = list(set(ds.coords).intersection(x_dims))[0]
+        x_dim = cast("str", x_dim)
     except IndexError:
         return None
     else:
@@ -314,3 +295,67 @@ def transform2tuple(transform: rio.Affine) -> tuple[float, float, float, float,
         The affine transform as a tuple (a, b, c, d, e, f)
     """
     return (transform.a, transform.b, transform.c, transform.d, transform.e, transform.f)
+
+
+def write_crs(ds: XD) -> XD:
+    """Write geo reference info into a dataset or dataarray."""
+    ds = ds.rio.write_transform()
+    if ds.rio.grid_mapping and ds.rio.grid_mapping != "spatial_ref":
+        ds = ds.rio.write_crs(ds.rio.crs, grid_mapping_name=ds.rio.grid_mapping)
+        if "spatial_ref" in ds:
+            ds = ds.drop_vars(["spatial_ref"])
+    else:
+        ds = ds.rio.write_crs(ds.rio.crs)
+    ds = ds.rio.write_coordinate_system()
+    return ds
+
+
+def get_gtiff_attrs(
+    resp: bytes,
+    ds_dims: tuple[str, str] | None = None,
+    driver: str | None = None,
+    nodata: NUMBER | None = None,
+) -> Attrs:
+    """Get nodata, CRS, and dimension names in (vertical, horizontal) order from raster in bytes.
+
+    Parameters
+    ----------
+    resp : bytes
+        Raster response returned from a wed service request such as WMS
+    ds_dims : tuple of str, optional
+        The names of the vertical and horizontal dimensions (in that order)
+        of the target dataset, default to None. If None, dimension names are determined
+        from a list of common names.
+    driver : str, optional
+        A GDAL driver for reading the content, defaults to automatic detection. A list of
+        the drivers can be found here: https://gdal.org/drivers/raster/index.html
+    nodata : float or int, optional
+        The nodata value of the raster, defaults to None, i.e., is determined from the raster.
+
+    Returns
+    -------
+    Attrs
+        No data, CRS, and dimension names for vertical and horizontal directions or
+        a list of the existing dimensions if they are not in a list of common names.
+    """
+    with rio.MemoryFile() as memfile:
+        memfile.write(resp)
+        with memfile.open(driver=driver) as src:
+            r_crs = pyproj.CRS(src.crs)
+            _nodata = get_nodata(src) if nodata is None else nodata
+
+            ds = rxr.open_rasterio(src)
+            ds = cast("xr.Dataset", ds)
+            if ds_dims is None:
+                ds_dims = get_dim_names(ds)
+            valid_dims = list(ds.sizes)
+            ds.close()
+
+            valid_dims = cast("list[str]", valid_dims)
+            if ds_dims is None or any(d not in valid_dims for d in ds_dims):
+                raise MissingAttributeError("ds_dims", valid_dims)
+            if isinstance(src.transform, rio.Affine):
+                transform = transform2tuple(src.transform)
+            else:
+                transform = tuple(src.transform)
+    return Attrs(_nodata, r_crs, ds_dims, transform)
