@@ -30,7 +30,7 @@ def netcdf_and_hdf5_versions() -> list[tuple[str, str | None]]:
         libhdf5_version = netCDF4.__hdf5libversion__
         libnetcdf_version = netCDF4.__netcdf4libversion__
     elif importlib.util.find_spec("h5py"):
-        import h5py
+        import h5py  # type: ignore
 
         libhdf5_version = h5py.version.hdf5_version
 
