@@ -4,42 +4,47 @@ from __future__ import annotations
 import itertools
 import tempfile
 import uuid
+from collections import defaultdict
 from dataclasses import dataclass
 from pathlib import Path
-from typing import TYPE_CHECKING, Any, Sequence, Tuple, TypeVar, Union
+from typing import TYPE_CHECKING, Any, Sequence, Tuple, TypeVar, Union, cast
 
-import cytoolz as tlz
-import dask
+import cytoolz.curried as tlz
+import dask.config
 import geopandas as gpd
 import numpy as np
+import numpy.typing as npt
 import pandas as pd
 import pyproj
-import rasterio as rio
+import rasterio.features as rio_features
 import rasterio.transform as rio_transform
 import rioxarray as rxr
 import shapely.geometry as sgeom
 import ujson as json
 import xarray as xr
+from pyproj.exceptions import CRSError as ProjCRSError
+from rasterio import MemoryFile
 from scipy.interpolate import BSpline
 from shapely import ops
 from shapely.geometry import LineString, MultiPolygon, Polygon
 
-from . import _utils as utils
-from ._utils import Attrs
-from .exceptions import (
+from pygeoutils import _utils as utils
+from pygeoutils.exceptions import (
     EmptyResponseError,
     InputRangeError,
     InputTypeError,
     InputValueError,
-    MissingAttributeError,
+    MatchingCRSError,
     MissingColumnError,
     MissingCRSError,
     UnprojectedCRSError,
 )
 
 BOX_ORD = "(west, south, east, north)"
-NUMBER = Union[int, float, np.number]
+NUMBER = Union[int, float, np.number]  # type: ignore
 if TYPE_CHECKING:
+    from rasterio import Affine
+
     GTYPE = Union[Polygon, MultiPolygon, Tuple[float, float, float, float]]
     GDF = TypeVar("GDF", gpd.GeoDataFrame, gpd.GeoSeries)
     XD = TypeVar("XD", xr.Dataset, xr.DataArray)
@@ -59,6 +64,7 @@ __all__ = [
     "geodf2xarray",
     "Coordinates",
     "GeoBSpline",
+    "query_indices",
     "nested_polygons",
 ]
 
@@ -128,58 +134,49 @@ def json2geodf(
         geodf = geodf.set_crs(in_crs)
         if in_crs != crs:
             geodf = geodf.to_crs(crs)
-
+    geodf = cast("gpd.GeoDataFrame", geodf)
     return geodf
 
 
 def geo2polygon(
     geometry: GTYPE,
-    geo_crs: CRSTYPE,
-    crs: CRSTYPE,
-) -> Polygon:
+    geo_crs: CRSTYPE | None = None,
+    crs: CRSTYPE | None = None,
+) -> Polygon | MultiPolygon:
     """Convert a geometry to a Shapely's Polygon and transform to any CRS.
 
     Parameters
     ----------
     geometry : Polygon or tuple of length 4
         Polygon or bounding box (west, south, east, north).
-    geo_crs : int, str, or pyproj.CRS
-        Spatial reference of the input geometry
+    geo_crs : int, str, or pyproj.CRS, optional
+        Spatial reference of the input geometry, defaults to ``None``.
     crs : int, str, or pyproj.CRS
-        Target spatial reference.
+        Target spatial reference, defaults to ``None``.
 
     Returns
     -------
-    Polygon
-        A Polygon in the target CRS.
+    shapely.Polygon or shapely.MultiPolygon
+        A (Multi)Polygon in the target CRS, if different from the input CRS.
     """
-    if not isinstance(geometry, (Polygon, MultiPolygon, Sequence)):
-        raise InputTypeError("geometry", "Polygon or tuple of length 4")
+    if isinstance(geometry, (Polygon, MultiPolygon)):
+        geom = geometry
+    elif isinstance(geometry, (tuple, list)) and len(geometry) == 4:
+        geom = sgeom.box(*geometry)
+    else:
+        raise InputTypeError("geometry", "(Multi)Polygon or tuple of length 4")
 
-    if isinstance(geometry, Sequence) and len(geometry) != 4:
-        raise InputTypeError("geometry", "tuple of length 4")
+    if geo_crs and crs and pyproj.CRS(geo_crs) != pyproj.CRS(crs):
+        project = pyproj.Transformer.from_crs(geo_crs, crs, always_xy=True).transform
+        geom = ops.transform(project, geom)
 
-    geom = sgeom.box(*geometry) if isinstance(geometry, Sequence) else geometry
-    project = pyproj.Transformer.from_crs(geo_crs, crs, always_xy=True).transform
-    geom = ops.transform(project, geom)
+    geom = cast("Polygon | MultiPolygon", geom)
     if not geom.is_valid:
         geom = geom.buffer(0.0)
+        geom = cast("Polygon | MultiPolygon", geom)
     return geom
 
 
-def _write_crs(ds: XD) -> XD:
-    """Write geo reference info into a dataset or dataarray."""
-    ds = ds.rio.write_transform()
-    if ds.rio.grid_mapping and ds.rio.grid_mapping != "spatial_ref":
-        ds = ds.rio.write_crs(ds.rio.crs, grid_mapping_name=ds.rio.grid_mapping)
-        if "spatial_ref" in ds:
-            ds = ds.drop_vars(["spatial_ref"])
-    else:
-        ds = ds.rio.write_crs(ds.rio.crs)
-    ds = ds.rio.write_coordinate_system()
-    return ds
-
-
 def xarray_geomask(
     ds: XD,
     geometry: GTYPE,
@@ -217,20 +214,19 @@ def xarray_geomask(
         The input dataset with a mask applied (np.nan)
     """
     ds_attrs = ds.attrs
-    if isinstance(ds, xr.Dataset):
-        da_attrs = {v: ds[v].attrs for v in ds}
+    da_attrs = {v: ds[v].attrs for v in ds} if isinstance(ds, xr.Dataset) else {}
 
     if ds.rio.crs is None:
         raise MissingCRSError
 
     geom = geo2polygon(geometry, crs, ds.rio.crs)
-    ds = _write_crs(ds)
+    ds = utils.write_crs(ds)
     ds = ds.rio.clip_box(*geom.bounds, auto_expand=True)
     if isinstance(geometry, (Polygon, MultiPolygon)):
         ds = ds.rio.clip([geom], all_touched=all_touched, drop=drop, from_disk=from_disk)
 
     if drop:
-        ds = _write_crs(ds)
+        ds = utils.write_crs(ds)
     ds.rio.update_attrs(ds_attrs, inplace=True)
     if isinstance(ds, xr.Dataset):
         _ = [ds[v].rio.update_attrs(da_attrs[v], inplace=True) for v in ds]
@@ -238,54 +234,6 @@ def xarray_geomask(
     return ds
 
 
-def get_gtiff_attrs(
-    resp: bytes,
-    ds_dims: tuple[str, str] | None = None,
-    driver: str | None = None,
-    nodata: NUMBER | None = None,
-) -> Attrs:
-    """Get nodata, CRS, and dimension names in (vertical, horizontal) order from raster in bytes.
-
-    Parameters
-    ----------
-    resp : bytes
-        Raster response returned from a wed service request such as WMS
-    ds_dims : tuple of str, optional
-        The names of the vertical and horizontal dimensions (in that order)
-        of the target dataset, default to None. If None, dimension names are determined
-        from a list of common names.
-    driver : str, optional
-        A GDAL driver for reading the content, defaults to automatic detection. A list of
-        the drivers can be found here: https://gdal.org/drivers/raster/index.html
-    nodata : float or int, optional
-        The nodata value of the raster, defaults to None, i.e., is determined from the raster.
-
-    Returns
-    -------
-    Attrs
-        No data, CRS, and dimension names for vertical and horizontal directions or
-        a list of the existing dimensions if they are not in a list of common names.
-    """
-    with rio.MemoryFile() as memfile:
-        memfile.write(resp)
-        with memfile.open(driver=driver) as src:
-            r_crs = pyproj.CRS(src.crs)
-            _nodata = utils.get_nodata(src) if nodata is None else nodata
-
-            ds = rxr.open_rasterio(src)
-            if ds_dims is None:
-                ds_dims = utils.get_dim_names(ds)
-
-            valid_dims = list(ds.sizes)
-            if ds_dims is None or any(d not in valid_dims for d in ds_dims):
-                raise MissingAttributeError("ds_dims", valid_dims)
-            if isinstance(src.transform, rio.Affine):
-                transform = utils.transform2tuple(src.transform)
-            else:
-                transform = tuple(src.transform)  # type: ignore
-    return Attrs(_nodata, r_crs, ds_dims, transform)
-
-
 def gtiff2xarray(
     r_dict: dict[str, bytes],
     geometry: GTYPE | None = None,
@@ -344,17 +292,18 @@ def gtiff2xarray(
     if "_dd_" in key1:
         var_name = {lyr: "_".join(lyr.split("_")[:-3]) for lyr in r_dict.keys()}
 
-    attrs = get_gtiff_attrs(r_dict[key1], ds_dims, driver, nodata)
+    attrs = utils.get_gtiff_attrs(r_dict[key1], ds_dims, driver, nodata)
     dtypes: dict[str, type] = {}
     nodata_dict: dict[str, NUMBER] = {}
 
     tmp_dir = tempfile.gettempdir()
 
     def to_dataset(lyr: str, resp: bytes) -> Path:
-        with rio.MemoryFile() as memfile:
+        with MemoryFile() as memfile:
             memfile.write(resp)
             with memfile.open(driver=driver) as vrt:
-                ds = rxr.open_rasterio(vrt)
+                ds = rxr.open_rasterio(vrt)  # type: ignore
+                ds = cast("xr.Dataset", ds)
                 if "band" in ds.dims:
                     ds = ds.squeeze("band", drop=True)
                 ds.name = var_name[lyr]
@@ -367,7 +316,7 @@ def gtiff2xarray(
 
     with dask.config.set(**{"array.slicing.split_large_chunks": True}):
         ds = xr.open_mfdataset(
-            itertools.starmap(to_dataset, r_dict.items()),
+            itertools.starmap(to_dataset, r_dict.items()),  # type: ignore
             chunks="auto",
             parallel=True,
             engine="rasterio",
@@ -377,13 +326,15 @@ def gtiff2xarray(
         ds = ds.squeeze("band", drop=True)
 
     variables = list(ds)
+    variables = cast("str", variables)
 
     if len(variables) == 1:
         ds = ds[variables[0]].copy()
         ds = ds.astype(dtypes[variables[0]])
+        name = cast("str", ds.name)
         ds.attrs["crs"] = attrs.crs.to_string()
-        ds.attrs["nodatavals"] = (nodata_dict[ds.name],)
-        ds = ds.rio.write_nodata(nodata_dict[ds.name])
+        ds.attrs["nodatavals"] = (nodata_dict[name],)
+        ds = ds.rio.write_nodata(nodata_dict[name])
     else:
         ds.attrs["crs"] = attrs.crs.to_string()
         for v in variables:
@@ -396,7 +347,7 @@ def gtiff2xarray(
     ds = ds.rio.write_crs(attrs.crs)
     ds = ds.rio.write_coordinate_system()
 
-    if isinstance(geometry, (Polygon, MultiPolygon)):
+    if geometry:
         if geo_crs is None:
             raise MissingCRSError
         return xarray_geomask(ds, geometry, geo_crs, all_touched, drop)
@@ -406,7 +357,7 @@ def gtiff2xarray(
 def get_transform(
     ds: xr.Dataset | xr.DataArray,
     ds_dims: tuple[str, str] = ("y", "x"),
-) -> tuple[rio.Affine, int, int]:
+) -> tuple[Affine, int, int]:
     """Get transform of a ``xarray.Dataset`` or ``xarray.DataArray``.
 
     Parameters
@@ -471,23 +422,23 @@ def xarray2geodf(
     valid_types = ["int16", "int32", "uint8", "uint16", "float32"]
     if dtype not in valid_types:
         raise InputValueError("dtype", valid_types)
-    _dtype = getattr(np, dtype)
 
     crs = da.rio.crs if da.attrs.get("crs") is None else da.crs
     if crs is None:
         raise MissingCRSError
 
     mask = None if mask_da is None else mask_da.to_numpy()
-    shapes = rio.features.shapes(
-        source=da.to_numpy().astype(_dtype),
+    shapes = rio_features.shapes(
+        source=da.to_numpy().astype(dtype),
         transform=da.rio.transform(),
         mask=mask,
         connectivity=connectivity,
     )
-    geometry, values = zip(*shapes)
+    geojsons, values = zip(*shapes)
+    geojsons = cast("tuple[dict[str, Any], ...]", geojsons)
     return gpd.GeoDataFrame(
-        data={da.name: _dtype(values)},
-        geometry=[sgeom.shape(g) for g in geometry],
+        data={str(da.name): np.array(values, dtype)},
+        geometry=[sgeom.shape(g) for g in geojsons],
         crs=crs,
     )
 
@@ -530,10 +481,11 @@ def geodf2xarray(
         raise InputTypeError("projected_crs", "a projected CRS")
 
     gdf = geodf.to_crs(projected_crs) if geodf.crs != pyproj.CRS(projected_crs) else geodf
+    gdf = cast("gpd.GeoDataFrame", gdf)
     west, south, east, north = gdf.total_bounds
     width = np.ceil(abs(west - east) / resolution).astype(int)
     height = np.ceil(abs(north - south) / resolution).astype(int)
-    affine = rio.transform.from_bounds(west, south, east, north, width, height)
+    affine = rio_transform.from_bounds(west, south, east, north, width, height)
 
     if attr_col:
         _types = ["int16", "int32", "uint8", "uint16", "uint32", "float32", "float64"]
@@ -543,7 +495,7 @@ def geodf2xarray(
             raise InputTypeError("attr_col", ", ".join(_types))
 
         ds = xr.DataArray(
-            rio.features.rasterize(
+            rio_features.rasterize(
                 shapes=zip(gdf.geometry, gdf[attr_col]),
                 out_shape=(height, width),
                 transform=affine,
@@ -556,7 +508,7 @@ def geodf2xarray(
         )
     else:
         ds = xr.DataArray(
-            rio.features.rasterize(
+            rio_features.rasterize(
                 shapes=gdf.geometry,
                 out_shape=(height, width),
                 transform=affine,
@@ -599,14 +551,15 @@ class Coordinates:
     @staticmethod
     def __box_geo(bounds: tuple[float, float, float, float] | None) -> sgeom.Polygon:
         """Get EPSG:4326 CRS."""
+        wgs84_bounds = pyproj.CRS(4326).area_of_use.bounds  # type: ignore
         if bounds is None:
-            return sgeom.box(*pyproj.CRS(4326).area_of_use.bounds)
+            return sgeom.box(*wgs84_bounds)
 
         if not isinstance(bounds, (tuple, list)) or len(bounds) != 4:
             raise InputTypeError("bounds", "tuple of length 4")
 
         bbox = sgeom.box(*bounds)
-        if not bbox.within(sgeom.box(*pyproj.CRS(4326).area_of_use.bounds)):
+        if not bbox.within(sgeom.box(*wgs84_bounds)):
             raise InputRangeError("bounds", "within EPSG:4326")
         return bbox
 
@@ -617,11 +570,19 @@ class Coordinates:
 
     def __post_init__(self) -> None:
         """Normalize the longitude value within [-180, 180)."""
-        _lon = [self.lon] if isinstance(self.lon, (int, float)) else list(self.lon)
-        lat = [self.lat] if isinstance(self.lat, (int, float)) else list(self.lat)
+        if isinstance(self.lon, (int, float, np.number)):
+            _lon = np.array([self.lon], "f8")
+        else:
+            _lon = np.array(self.lon, "f8")
+
+        if isinstance(self.lat, (int, float, np.number)):
+            lat = np.array([self.lat], "f8")
+        else:
+            lat = np.array(self.lat, "f8")
 
         lon = np.mod(np.mod(_lon, 360.0) + 540.0, 360.0) - 180.0
-        pts = gpd.GeoSeries([sgeom.Point(xy) for xy in zip(lon, lat)], crs=4326)
+        pts = gpd.GeoSeries([sgeom.Point(xy) for xy in zip(lon, lat)])
+        pts = cast("gpd.GeoSeries", pts.set_crs(4326))
         self._points = self.__validate(pts, self.__box_geo(self.bounds))
 
     @property
@@ -645,7 +606,7 @@ def validate_crs(crs: CRSTYPE) -> str:
     """
     try:
         return pyproj.CRS(crs).to_string()  # type: ignore
-    except pyproj.exceptions.CRSError as ex:
+    except ProjCRSError as ex:
         raise InputTypeError("crs", "a valid CRS") from ex
 
 
@@ -667,11 +628,11 @@ class Spline:
         Total distance of each point along the B-spline from the start point.
     """
 
-    x: np.ndarray
-    y: np.ndarray
-    phi: np.ndarray
-    radius: np.ndarray
-    distance: np.ndarray
+    x: npt.NDArray[np.float64]
+    y: npt.NDArray[np.float64]
+    phi: npt.NDArray[np.float64]
+    radius: npt.NDArray[np.float64]
+    distance: npt.NDArray[np.float64]
 
 
 class GeoBSpline:
@@ -713,8 +674,8 @@ class GeoBSpline:
 
     @staticmethod
     def __curvature(
-        xs: Sequence[float] | np.ndarray, ys: Sequence[float] | np.ndarray, l_tot: float
-    ) -> tuple[np.ndarray, np.ndarray]:
+        xs: npt.NDArray[np.float64], ys: npt.NDArray[np.float64], l_tot: float
+    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
         """Compute the curvature of a B-spline curve.
 
         Notes
@@ -778,12 +739,14 @@ class GeoBSpline:
         spl = BSpline(konts, np.column_stack([self.x_ln, self.y_ln]), degree)
 
         x_sp, y_sp = spl(np.linspace(0, self.npts_ln - degree, max(npts_sp, 3), endpoint=False)).T
+        x_sp = cast("npt.NDArray[np.float64]", x_sp)
+        y_sp = cast("npt.NDArray[np.float64]", y_sp)
         phi_sp, rad_sp = self.__curvature(x_sp, y_sp, self.l_ln)
         geom = (
             LineString([(x1, y1), (x2, y2)])
             for x1, y1, x2, y2 in zip(x_sp[:-1], y_sp[:-1], x_sp[1:], y_sp[1:])
         )
-        d_sp = gpd.GeoSeries(geom, crs=self.crs).length.cumsum().values
+        d_sp = gpd.GeoSeries(geom).set_crs(self.crs).length.cumsum().values
         if npts_sp < 3:
             idx = np.r_[:npts_sp]
             return Spline(x_sp[idx], y_sp[idx], phi_sp[idx], rad_sp[idx], d_sp[idx])
@@ -845,17 +808,18 @@ def snap2nearest(lines: GDF, points: GDF, tol: float) -> GDF:
         raise UnprojectedCRSError
 
     if isinstance(points, gpd.GeoSeries):
-        pts: gpd.GeoDataFrame = points.to_frame("geometry").reset_index()
+        pts = points.to_frame("geometry").reset_index()
     else:
         pts = points.copy()
 
+    pts = cast("gpd.GeoDataFrame", pts)
     cols = list(pts.columns)
     cols.remove("geometry")
     pts_idx, ln_idx = lines.sindex.query_bulk(pts.buffer(tol))
     merged_idx = tlz.merge_with(list, ({p: f} for p, f in zip(pts_idx, ln_idx)))
     _pts = {
         pi: (
-            *pts.iloc[pi][cols],  # type: ignore[has-type]
+            *pts.iloc[pi][cols],
             ops.nearest_points(lines.iloc[fi].geometry.unary_union, pts.iloc[pi].geometry)[0],
         )
         for pi, fi in merged_idx.items()
@@ -863,6 +827,7 @@ def snap2nearest(lines: GDF, points: GDF, tol: float) -> GDF:
     pts = gpd.GeoDataFrame.from_dict(_pts, orient="index")
     pts.columns = cols + ["geometry"]
     pts = pts.set_geometry("geometry", crs=points.crs)
+    pts = cast("gpd.GeoDataFrame", pts)
 
     if isinstance(points, gpd.GeoSeries):
         return pts.geometry
@@ -893,6 +858,9 @@ def break_lines(lines: GDF, points: gpd.GeoDataFrame, tol: float = 0.0) -> GDF:
     if lines.crs is None or points.crs is None:
         raise MissingCRSError
 
+    if lines.crs != points.crs or not lines.crs.is_projected or not points.crs.is_projected:
+        raise UnprojectedCRSError
+
     if "direction" not in points.columns:
         raise MissingColumnError(["direction"])
 
@@ -902,31 +870,23 @@ def break_lines(lines: GDF, points: gpd.GeoDataFrame, tol: float = 0.0) -> GDF:
     if not lines.geom_type.isin(["LineString", "MultiLineString"]).all():
         raise InputTypeError("geometry", "LineString or MultiLineString")
 
-    if lines.crs != points.crs or not lines.crs.is_projected or not points.crs.is_projected:
-        crs_proj = "epsg:3857"
-        lns = lines.to_crs(crs_proj)
-        pts = points.to_crs(crs_proj)
-    else:
-        crs_proj = lines.crs
-        lns = lines.copy()
-        pts = points.copy()
-
+    crs_proj = lines.crs
     if tol > 0.0:
-        pts = snap2nearest(lns, pts, tol)
+        points = snap2nearest(lines, points, tol)
 
-    mlines = lns.geom_type == "MultiLineString"
+    mlines = lines.geom_type == "MultiLineString"
     if mlines.any():
-        lns.loc[mlines, "geometry"] = lns.loc[mlines, "geometry"].apply(lambda g: list(g.geoms))
-        lns = lns.explode("geometry").set_crs(crs_proj)
+        lines.loc[mlines, "geometry"] = lines.loc[mlines, "geometry"].apply(lambda g: list(g.geoms))
+        lines = lines.explode("geometry").set_crs(crs_proj)
 
-    pts_idx, flw_idx = lns.sindex.query_bulk(pts.geometry)
+    pts_idx, flw_idx = lines.sindex.query_bulk(points.geometry)
     if len(pts_idx) == 0:
         raise ValueError("No intersection between lines and points")  # noqa: TC003
 
-    flw_geom = lns.iloc[flw_idx].geometry
-    pts_geom = pts.iloc[pts_idx].geometry
-    pts_dir = pts.iloc[pts_idx].direction
-    idx = lns.iloc[flw_idx].index
+    flw_geom = lines.iloc[flw_idx].geometry
+    pts_geom = points.iloc[pts_idx].geometry
+    pts_dir = points.iloc[pts_idx].direction
+    idx = lines.iloc[flw_idx].index
     broken_lines = gpd.GeoSeries(
         [
             ops.substring(fl, *((0, fl.project(pt)) if d == "up" else (fl.project(pt), fl.length)))
@@ -935,7 +895,7 @@ def break_lines(lines: GDF, points: gpd.GeoDataFrame, tol: float = 0.0) -> GDF:
         crs=crs_proj,
         index=idx,
     )
-    out = lns.loc[idx].drop(columns="geometry")
+    out = lines.loc[idx].drop(columns="geometry")
     out = gpd.GeoDataFrame(out, geometry=broken_lines, crs=crs_proj)
     return out.to_crs(lines.crs)
 
@@ -948,11 +908,11 @@ def geometry_list(
         return [geometry]
 
     if isinstance(geometry, (sgeom.MultiPolygon, sgeom.MultiLineString, sgeom.MultiPoint)):
-        return list(geometry.geoms)
+        return list(geometry.geoms)  # type: ignore
 
     if isinstance(geometry, (tuple, list)) and len(geometry) == 4:
         return [sgeom.box(*geometry)]
-    valid_geoms = [
+    valid_geoms = (
         "Polygon",
         "MultiPolygon",
         "tuple/list of length 4",
@@ -960,10 +920,43 @@ def geometry_list(
         "MultiPoint",
         "LineString",
         "MultiLineString",
-    ]
+    )
     raise InputTypeError("geometry", ", ".join(valid_geoms))
 
 
+def query_indices(
+    tree_gdf: gpd.GeoDataFrame | gpd.GeoSeries,
+    input_gdf: gpd.GeoDataFrame | gpd.GeoSeries,
+    predicate: str = "intersects",
+) -> dict[Any, list[Any]]:
+    """Find the indices of the input_geo that intersect with the tree_geo.
+
+    Parameters
+    ----------
+    tree_gdf : geopandas.GeoDataFrame or geopandas.GeoSeries
+        The tree geodataframe.
+    input_gdf : geopandas.GeoDataFrame or geopandas.GeoSeries
+        The input geodataframe.
+    predicate : str, optional
+        The predicate to use for the query operation, defaults to ``intesects``.
+
+    Returns
+    -------
+    dict
+        A dictionary of the indices of the ``input_gdf`` that intersect with the
+        ``tree_gdf``. Keys are the index of ``input_gdf`` and values are a list
+        of indices of the intersecting ``tree_gdf``.
+    """
+    if input_gdf.crs != tree_gdf.crs:
+        raise MatchingCRSError
+
+    in_iloc, tr_iloc = tree_gdf.sindex.query_bulk(input_gdf.geometry, predicate=predicate)
+    idx_dict = defaultdict(set)
+    for ii, it in zip(input_gdf.iloc[in_iloc].index, tree_gdf.iloc[tr_iloc].index):
+        idx_dict[ii].add(it)
+    return {k: list(v) for k, v in idx_dict.items()}
+
+
 def nested_polygons(gdf: gpd.GeoDataFrame | gpd.GeoSeries) -> dict[int | str, list[int | str]]:
     """Get nested polygons in a GeoDataFrame.
 
@@ -979,12 +972,18 @@ def nested_polygons(gdf: gpd.GeoDataFrame | gpd.GeoSeries) -> dict[int | str, li
         values are a list of indices of smaller polygons that are
         contained within the larger polygons.
     """
-    if gdf.geom_type.str.contains("Polygon").sum() != len(gdf):
+    if not gdf.geom_type.str.contains("Polygon").all():
         raise InputTypeError("gdf", "dataframe with (Multi)Polygons")
 
-    gdf = gdf.to_frame() if isinstance(gdf, gpd.GeoSeries) else gdf
-    q_idx, t_idx = gdf.sindex.query_bulk(gdf.geometry, predicate="within")
-    merged_idx = tlz.merge_with(
-        list, ({gdf.iloc[t].name: gdf.iloc[q].name} for q, t in zip(q_idx, t_idx) if t != q)
-    )
-    return merged_idx  # type: ignore[no-any-return]
+    if gdf.crs is None or not gdf.crs.is_projected:
+        raise UnprojectedCRSError
+
+    centroid = gdf.centroid
+    nested_idx = query_indices(centroid, gdf, "contains")
+    nested_idx = {k: list(set(v).difference({k})) for k, v in nested_idx.items()}
+    nested_idx = {k: v for k, v in nested_idx.items() if v}
+    nidx = {tuple(set(v + [k])) for k, v in nested_idx.items()}
+    area = gdf.area
+    nested_keys = [area.loc[list(i)].idxmax() for i in nidx]
+    nested_idx = {k: v for k, v in nested_idx.items() if k in nested_keys}
+    return nested_idx
