@@ -39,7 +39,7 @@ def test_geom_list():
 
 
 def test_break_line():
-    crs_proj = "epsg:3857"
+    crs_proj = 3857
     lines = gpd.GeoSeries([LineString([[0, 0], [2, 2]])], crs=crs_proj)
     pt = Point(1, 1)
     points = gpd.GeoDataFrame({"direction": ["down"]}, geometry=[pt], crs=crs_proj, index=[0])
@@ -353,7 +353,8 @@ def test_path():
 
 def test_nested():
     gdf = gpd.GeoSeries(
-        [box(6, 6, 8, 8), box(3, 3, 4, 4), box(2, 2, 5, 5), box(5, 5, 6, 6), box(6, 6, 7, 7)]
+        [box(6, 6, 8, 8), box(3, 3, 4, 4), box(2, 2, 5, 5), box(5, 5, 6, 6), box(6, 6, 7, 7)],
+        crs=5070,
     )
     assert geoutils.nested_polygons(gdf) == {2: [1], 0: [4]}
 
